;;;
;;;    Help and Apropos    by T.Komiya
;;;      usage: (help) or (help '<topics>)
;;;             (apropos "<str>")
;;;

(macro (when form) `(if ,(cadr form) (begin ,@(cddr form))))

;;;

(define (init-help fname)
  (let ((port (open-input-file fname)))
    (with-input-from-port port
      (lambda ()
        (do ((entry (read) (read)))
            ((eof-object? entry))
          (db 'set-help (car entry) entry))))))

(define (help . subject)
  (if (null? subject)
      (db 'show-usage)
      (db 'fetch-help (car subject))))

(define (apropos str)
  (db 'apropos str))

(define db
  (let ((topics '()))
    (define (show-usage)
      (display "Usage: (help '<topics>)")
      (newline)
      (display "       <topics> must be procedure, syntactic keyword or macro.")
      (newline) (newline)
      (display "       (apropos <str>)")
      (newline)
      (display "       Prints symbols whose print-names contain <str> as substring.")
      (newline))
    (define (fetch-help subject)
      (let ((desc (getprop 'topics subject)))
	(if (not desc)
	    (format #t "No documentation for ~a~%" subject)
	    (const-desc desc))))
    (define (set-help subject desc)
      (putprop 'topics subject desc))
    (define (show-topics topics-list)
      (if (not (pair? topics-list))
	  '()
          (begin
	    (format #t "~a~%" (caar topics-list))
	    (show-topics (cdr topics-list)))))
    (define (apropos-help str topics-list)
      (if (not (pair? topics-list))
	  #f
          (begin
	    (if (substring? str (symbol->string (caar topics-list)))
		(format #t "~a~%" (caar topics-list)))
	    (apropos-help str (cdr topics-list)))))
    (lambda (msg . args)
      (case msg
        ((show-usage) (show-usage))
        ((fetch-help) (fetch-help (car args)))
	((set-help) (set-help (car args) (cadr args)))
	((show-topics) (show-topics (symbol-plist 'topics)))
	((apropos) (apropos-help (car args) (symbol-plist 'topics)))
	(else (error "Unknown message ~a." meg))))))

(define (const-desc desc)
  (let* ((subj (car desc))
         (strsubj (symbol->string subj))
         (arginfo (cadr desc)))
    (display "---------------------------------------------------------------------------")
    (newline)
    (display "Syntax: ")
    (display (make-string 56 #\space))
    (display (obj-class subj)) (newline)
    (if (zero? (string-length arginfo))
	(format #t "  (~a)~%~%" subj)
        (format #t "  (~a ~a)~%~%" subj arginfo))
    (do ((d (cddr desc) (cdr d)))
	((null? d))
      (display (car d)) (newline))
    (display "--------------------------------------------------------------------------")
    '-))

(define (obj-class subj)
  (if (macro-function subj)
      '[macro]
      (if (bound? subj)
          (let ((value (eval subj)))
            (if (function? value)
                '[procedure]
                '[special]))
          '[special])))

(define (substring? sub str)
  (do ((i (- (string-length str) (string-length sub)) i)
       (l (string-length sub) l)
       (j 0 (1+ j))
       (ret #f ret))
      ((> j i) ret)
      (when (string=? sub (substring str j (+ j l)))
	    (set! ret #t)
	    (set! j (1+ i)))))

(init-help (string-append *faslib-path* "/help.dat"))
